/*******************************************************************************


   VOLTMETER_AUTO_LED_C

	PROJETO: VOLTIMETRO COM 3 DIGITOS PARA USO EM VECULOS (12V)

   DATA: 17/07/2016

	AUTOR:  CLUDIO LRIOS
	PROCESSADOR:  PIC   16F628A
	OBJETIVO: MONITORAR A TENSO DA BATERIA DO VECULO. CASO A TENSO DA BATERIA
   FIQUE ABAIXO DA TENSO MINIMA, DESLIGA UM RELE (O QUAL PODER DESLIGAR
   SAIDA DE SOM AUTOMOTIVO).
   ACRESCENTADO UM LED PARA INDICAR SOBRECARGA DO ALTERNADOR >14,5 VOLTS

   UTILIZA O METODO DELTA_SIGMA PARA CONVERTER A TENSO ANALGICA EM DIGITAL.
   (MULTIPLEXAAO E BASE DE TEMPO PARA O CONVERSOR USANDO INTERRUPO DO TIMER0)

   PINOS DO MICROCONTROLADOR:
   1- NC
   2- SEGUIMENTO 'C'
   3- SEGUIMENTO 'D'
   4- RESET
   5- GND
   6- SEGUIMENTO 'A'
   7- SEGUIMENTO 'B'
   8- CATODO COMUM DISPLAY '0'
   9- NC
   10-CATODO COMUM DISPLAY '2'
   11-SEGUIMENTO 'E'
   12-CATODO COMUM DISPLAY '1' E PONTO
   13-NC
   14-VCC
   15-SEGUIMENTO 'F'
   16-SEGUIMENTO 'G'
   17-RESISTOR DE 47K (DESCARGA)
   18-RESISTOR DE 150 (ENTRADA)

   Use os 2 resistores de preciso de 47k 1% na entrada do conversor.
   Capacitor de 220nF poliester.
   Este arquivo  parte integrante do BLOG LARIOS.TECNOLOGIA.WS.
   Uso para fins didticos.
********************************************************************************/

      #DEFINE  DISPLAY_CATODO_COMUM

// comente o acima para anodo comum e descomente para catodo comum
//===============================================================================
//

 #define TENSAO_LIMITE  1100    //  11,00 Volts como tenso de desligamento
 #define VALOR_FILTRO   5       //valor p/ filtragem de picos abaixo da tenso
 #define TENSAO_LIMITE_MAX 1450 // 14,5 Volts. Se acima haver excesso de carga
 
 
 
//==============================================================================

#include <16F628A.h>
#use delay(clock=4000000)//OSCILADOR RC INTERNO DO PIC
#fuses NOWDT,INTRC_IO, NOPUT, NOPROTECT, BROWNOUT, MCLR, NOLVP, NOCPD
#use fast_io(a)//MODO RPIDO
#use fast_io(b)//MODO RPIDO
#zero_ram  // ZERA RAM INICIALMENTE
//==============================================================================
//   Bytes
//==============================================================================

#byte trisa = 0x85
#byte trisb = 0x86
#byte porta = 0x05
#byte portb = 0x06
#byte cmcon = 0x1f
#byte vrcon = 0x9f
#byte timer0= 0x01
//==============================================================================
//   Bits
//==============================================================================

#bit  k0    = 0x06.2 //k0
#bit  k1    = 0x06.6 //k1
#bit  k2    = 0x06.4 //k2
#bit  desc  = 0x05.0 //descarga ra0
#bit  c2_out= 0x1f.7// cmcon saida c2
#bit  seg_a = 0x06.0 //seg a
#bit  seg_b = 0x06.1 //seg b
#bit  seg_c = 0x05.3 //seg c
#bit  seg_d = 0x05.4 //seg d
#bit  seg_e = 0x06.5 //seg e
#bit  seg_f = 0x05.6 //seg f
#bit  seg_g = 0x05.7 //seg g

#bit  rele  = 0x06.7  //sada para o rele
#bit  L_D   = 0x06.3  //entrada de liga/desliga display
#bit  led   = 0x06.3  //sada de led de sobrecarga
#bit  tris_led = 0x86.3 //sentido do port
//==============================================================================
//   Variveis globais
//==============================================================================

int16 vbat=0;
int8 resc0,resc1,cont_f=VALOR_FILTRO;
int1   flag_ds,f_desc,f_ld=0,f_led=0;

//==============================================================================
//   Array com valor em 7 seguimentos
//   (obs. modo catodo comum, precisando inverte ao usar para anodo comum)
//==============================================================================
const int8 c7seg [10]={
//        (ABCDEFGP)
			0b11111100,	 /* RETORNA 0    FC */
			0b01100000,  /* RETORNA 1    60 */
			0b11011010,  /* RETORNA 2	  DA */
			0b11110010,	 /* RETORNA 3	  F2 */
			0b01100110,	 /* RETORNA 4	  66 */
			0b10110110,	 /* RETORNA 5	  B6 */
			0b10111110,	 /* RETORNA 6	  BE */
			0b11100000,	 /* RETORNA 7	  E0 */
			0b11111110,	 /* RETORNA 8    FE */
			0b11110110	 /* RETORNA 9    F6 */
};
//==============================================================================
//   //Estrutura para fins de deslocamento de dados
//==============================================================================
 struct to_shift {
  int8 c0;
  int8 c1;
  int8 res0;
  int8 res1;
  int8 res2;
 } res;

//==============================================================================
//   Rotina de envio para sada dos seguimentos
//==============================================================================
void env_out(int8 a){
#IFDEF  DISPLAY_CATODO_COMUM
 if(bit_test(a,7))  seg_a=1; else seg_a=0;
 if(bit_test(a,6))  seg_b=1; else seg_b=0;
 if(bit_test(a,5))  seg_c=1; else seg_c=0;
 if(bit_test(a,4))  seg_d=1; else seg_d=0;
 if(bit_test(a,3))  seg_e=1; else seg_e=0;
 if(bit_test(a,2))  seg_f=1; else seg_f=0;
 if(bit_test(a,1))  seg_g=1; else seg_g=0;
 #ELSE
 if(bit_test(a,7))  seg_a=0; else seg_a=1;
 if(bit_test(a,6))  seg_b=0; else seg_b=1;
 if(bit_test(a,5))  seg_c=0; else seg_c=1;
 if(bit_test(a,4))  seg_d=0; else seg_d=1;
 if(bit_test(a,3))  seg_e=0; else seg_e=1;
 if(bit_test(a,2))  seg_f=0; else seg_f=1;
 if(bit_test(a,1))  seg_g=0; else seg_g=1;

 #ENDIF
}

//==============================================================================
//   Rotina de interrupo do timer 0
//==============================================================================
#INT_TIMER0
void int_tmr0_isr(){
  static int8 vez,vez1;
  int8 aux;

  timer0=155;//carga do timer para 100 useg
  flag_ds=1;//sinaliza para o conversor delta_sigma

  if(++vez1>32) { // verifica se deve mudar digito do display (multiplexao)
  vez1=0;//zera contador auxiliar de multiplexao
  k0=0;//desliga anodos (apagar displays)
  k1=0;
  k2=0;

  if(!f_ld) return;//se no recebe sinal do remoto desliga display e sai da int.

  if(++vez>2) vez=0;//incrementa e limita a 3 digitos
  switch (vez){
    case 0: {//seleciona digito '1'(lsb) para apresentar
      aux=resc0;//pega valor
      aux=swap(aux)&0x0f; //inverte e limita a um nibble
      aux=c7seg[aux];//converte para 7 seguimentos
      env_out(aux); //envia para sadas dos seg.
      k0=1;//liga anodo do 1 digito
      break;
  }

    case 1: {//seleciona digito '2' para apresentar
      aux=resc1&0x0f;
      aux=c7seg[aux];
      env_out(aux);//envia para sadas dos seg.
      k1=1;//liga anodo do 2 digito
      break;
  }
    case 2: {//seleciona digito '3' (msb) para apresentar
      aux=resc1;
      aux=swap(aux)&0x0f;
      if(!aux)break;
      aux=c7seg[aux];
      env_out(aux);//envia para sadas dos seg.
      k2=1;//liga anodo do 3 digito
      break;
  }

  }
 }
}
//==============================================================================
//   Rotina do conversor Delta Sigma
//   obs. usa o timer 0 para produzir tempos iguais entre ciclos
//==============================================================================
int16 get_adc_ds(){
  int16 ciclo=0x1000; //12 bits de resoluo
  int16 val=0;//zera resultado

  for(;;){
   while(!flag_ds);//aguarda momento correto
   flag_ds=0;//zera flag

   if(c2_out){desc=1; } else {desc=0; val++;}//testa sada do comparador
   if(!--ciclo){desc=0; vbat=val;return(val); }//testa fim da converso adc delta-sigma
 }
}
//==============================================================================
//Converte hex 2 bytes para decimal 3 bytes
//==============================================================================
void bin2_dec(int16 a){
  int8  cont,aux;

  res.c0=make8(a,0);//carrega valor a converter
  res.c1=make8(a,1);
  res.res0=0;//zera resultados
  res.res1=0;
  res.res2=0;
  cont=16;//numero de vezes da converso
 for(;;){//loop continuo
  shift_left(&res,5,0);//desloca 5 bytes p/ esquerda
  if(!--cont){//testa se chegou ao fim da converso
   resc0=res.res0;//copia para o buffer de apresentao
   resc1=res.res1;
   return;//fim da converso
  }
  aux=res.res2;
  if((aux+0x03)&0x08) res.res2+=0x03;//testa bit 3
  aux=res.res2;
  if((aux+0x30)&0x80) res.res2+=0x30;//testa bit 7
   aux=res.res1;
  if((aux+0x03)&0x08) res.res1+=0x03;//testa bit 3
  aux=res.res1;
  if((aux+0x30)&0x80) res.res1+=0x30;//testa bit 7
   aux=res.res0;
  if((aux+0x03)&0x08) res.res0+=0x03;//testa bit 3
  aux=res.res0;
  if((aux+0x30)&0x80) res.res0+=0x30;//testa bit 7
 }
}

//==============================================================================
//Testa entrada de comando que inibe o mostrador
//==============================================================================


void  testa_on_off(){
  tris_led=1; //entrada
  delay_us(50);//delay

  if(L_D ) {
  if(!f_desc) {rele=1;}else{ rele=0;} //uma vez desligado por falta de tenso, precisa desligar remote e ligar de novo
  f_ld=1;
  }
  else {rele=0; f_desc=0; f_ld=0;}
  tris_led=0;//sada



  //falta o tempo que ficar abaixo da tenso para desligar e bloquear o rele
}

//==============================================================================
//Verifica a tenso da bateria e compara com valor limite
//==============================================================================
void testa_tensao(){

  if((vbat<=TENSAO_LIMITE)&&(!f_desc)){
   if(--cont_f==0){ f_desc=1;}
   }
   else{
    cont_f=VALOR_FILTRO;//reinicia com valor
    
   }

}

//==============================================================================
//Verifica a tenso da bateria e compara com valor limite de sobrecarga do alternador
//==============================================================================


void testa_tensao_sobre_carga(){

 if(vbat>=TENSAO_LIMITE_MAX){ f_led=1; }
 else{ f_led=0;} 
}


//==============================================================================
//Indica por meio do led se est havendo situao de sobrecarga
//==============================================================================


void indica_sobrecarga(){

 if(f_led)led=1; else led=0;//aciona led de sobrecarga
}

//==============================================================================
//                  Rotina Principal
//==============================================================================

void main() {


   setup_counters(RTCC_INTERNAL,RTCC_DIV_1);
   setup_timer_1(T1_DISABLED);
   vrcon=0b11101100;//ajusta vref para 2,5 volts
   cmcon=5;//usando um comparador apenas
   trisa=0b00100110;
   portb=0;
   trisb=0b00000000;//TUDO SADA
   enable_interrupts(INT_TIMER0);
   enable_interrupts(GLOBAL);
   port_b_pullups(FALSE);
   f_desc=0;

//==============================================================================
//                 Loop Principal
//==============================================================================

   for(;;){

   bin2_dec(get_adc_ds());// obtem converso delta-sigma/ converte para hex e
                          // apresenta nos displays de 7 seg. multiplexados
   testa_tensao_sobre_carga(); //testa sobrecarga do alternador
   testa_on_off(); //verifica se foi ligado a entrada ('remote' do radio)
   testa_tensao(); //verifica se deve desligar rele.
   indica_sobrecarga();//aciona led de sobrecarga
   }
}
//==============================================================================
//                     Fim do Programa
//==============================================================================
